local handler = CreateFrame("ScrollFrame", "ObjectiveHandler", UIParent)
handler:SetPoint("TopLeft", MinimapCluster, "BottomRight")
function handler:InitializeController(parent, conf)
	self:SetController(parent)
	self:SetConfig(conf)
    self:Initialize(parent, conf)
    return self
end

function handler:SetController(parent)
	self:SetParent(parent)
	self:ClearAllPoints()
	self:SetPoint('TopLeft', 0, 0)
	self:SetPoint('BottomRight', 0, 0)
	self.parent = parent
end

function handler:SetConfig(conf)
	self.sets = conf
end

function handler:Initialize()
    self:SnatchObjectiveTrackerFrame()
    self:Hook_SetBlockHeader()
    self:HookMinimizeButton()
	self:InitializeScroll()
	self:CreateMenu()
    self:RegisterEvent("ZONE_CHANGED")
    self:RegisterEvent("ZONE_CHANGED_NEW_AREA")
    self:RegisterEvent("ZONE_CHANGED_INDOORS")
    self:RegisterEvent("WORLD_MAP_UPDATE")
    self:SetScript("OnEvent", handler.UpdateList)
	return self
end

function handler:SnatchObjectiveTrackerFrame()
    --Take initial control of the ObjectiveTrackerFrame
    self.tracker = ObjectiveTrackerFrame
	self.tracker:SetParent(self)
	self.tracker:SetMovable(true)
	self.tracker:SetUserPlaced(true)
	self.tracker:ClearAllPoints()
	self:SetScrollChild(self.tracker)
	self.tracker:SetPoint('TopLeft', 30, 0)
	self.tracker:SetHeight(2000)
end

function handler:HookMinimizeButton()
    --Toggle display of quest. Leaves tracked Achievements(etc.) visible.
	local mini = self.tracker.HeaderMenu.MinimizeButton
	mini:ClearAllPoints()
	mini:SetPoint("TopRight", self)
	local minimizeHandler = CreateFrame("Frame", nil, mini)
	minimizeHandler:SetParent(mini)
	minimizeHandler:SetAllPoints(mini)
	minimizeHandler:EnableMouse("AnyUp")
	minimizeHandler:SetScript("OnMouseDown", function(self, button)
		if button == "RightButton" then
			handler:ToggleQuests()
		else
			mini:Click()
		end
		ObjectiveTracker_Update(OBJECTIVE_TRACKER_UPDATE_QUEST)
	end)
end

function handler:Hook_SetBlockHeader()
    --corrects bug where all quests display an extra empty line of text.
	hooksecurefunc(QUEST_TRACKER_MODULE, "SetBlockHeader", function() self:RepositionTracker()
		for i = 1, GetNumQuestWatches() do
			local questID = GetQuestWatchInfo(i);
			if ( questID ) then
				local block = QUEST_TRACKER_MODULE:GetExistingBlock(questID);
				if ( block ) then
					for i, line in pairs(block.lines) do
						line.Text:SetHeight(28)
					end
				end
			end
		end
	 end)
end

function handler:CreateMenu()
    --drop down menu for quick acces to all quest sorting options.
	local quest = QUEST_TRACKER_MODULE.Header
	local header = CreateFrame("Frame", "QuestHeaderOverlay", quest)
	header:SetPoint("TopLeft", handler)
	header:SetPoint("TopRight", handler)
	header:SetHeight(25)
	quest:EnableMouse("RIGHTBUTTON")
	local dropDown = CreateFrame("Frame", "QuestFilterMenu", quest, "UIDropDownMenuTemplate")
	UIDropDownMenu_Initialize( dropDown, self.InitializeMenu, "MENU")

	quest:SetScript("OnMouseDown", function(self, ...)
		ToggleDropDownMenu(1, 1, dropDown, header:GetName(), 0, -5)
	end)
end

function handler:InitializeMenu(level)
	if level == 1 then
		local info = UIDropDownMenu_CreateInfo();
		info.text = "Filter By Zone";
		info.func = handler.ToggleFilterByZone;
		info.checked = handler.GetFilterByZone
		info.keepShownOnClick  = true
		UIDropDownMenu_AddButton(info);


		local info = UIDropDownMenu_CreateInfo();
		info.text = "Sort By Proximity";
		info.func = handler.ToggleSortByProximity;
		info.checked = handler.GetSortByProximity
		info.keepShownOnClick  = true
		UIDropDownMenu_AddButton(info);


		local info = UIDropDownMenu_CreateInfo();
		info.text = "Enable Scrolling";
		info.func = handler.ToggleScroll;
		info.checked = handler.GetScroll
		info.keepShownOnClick  = true
		UIDropDownMenu_AddButton(info);


		local info = UIDropDownMenu_CreateInfo();
		info.text = "Button on Left";
		info.func = handler.ToggleAlignment;
		info.checked = handler.GetAlignment
		info.keepShownOnClick  = true
		UIDropDownMenu_AddButton(info);


		local info = UIDropDownMenu_CreateInfo();
		info.text = "Max Quests Displayed";
		info.value = "max"
		info.notCheckable = true
		info.keepShownOnClick  = true
		info.hasArrow = 1
		UIDropDownMenu_AddButton(info);
	elseif level == 2 then
		for i = 1, 25 do
			local info = UIDropDownMenu_CreateInfo();
			info.text = i;
			info.func = function() handler:SetMaxDisplayed(i) end
			info.checked = function() return (handler:GetMaxDisplayed() == i) end
			UIDropDownMenu_AddButton(info, 2);
		end
	end
end

function handler:InitializeScroll()
    --Enables use of scrolling!
	self.maxRange = 1
	self:SetScript("OnMouseWheel", function(self, step)
		self:SetScroll((self:GetVerticalScroll() + (self:GetSettings().scrollStep * (step * -1))))
	end)
	self:SetVerticalScroll(-1)
	self:SetHorizontalScroll(-30)
	self.tracker:HookScript("OnEvent", function()
		self:SetScroll()
	end)
	self:SetScroll()
end

function handler:SetScroll(value)
	if not handler:GetScroll() then
		return
	end
	value = value or (self:GetVerticalScroll())
	local height = self:GetHeight()
	ObjectiveTrackerBlocksFrame.contentsHeight = ObjectiveTrackerBlocksFrame.contentsHeight or self:GetHeight()
	local min, max = -1, ceil(ObjectiveTrackerBlocksFrame.contentsHeight - height)
	if max < min then
		max = min
	end
	if (value < min)then
		value = min
	elseif value > max then
		value = max 
	end
	self:SetVerticalScroll(value)
	self:SetHorizontalScroll(-30)
end

function handler:UpdateItemButtons()
    --allows item buttons to be displayed on left side of ObjectiveTrackerFrame
	local onLeft = self:GetAlignment()
	for i = 1, GetNumQuestWatches() do
		local questID = GetQuestWatchInfo(i);
		if ( questID ) then
			local block = QUEST_TRACKER_MODULE:GetExistingBlock(questID);
			if ( block ) then
				if ( block.itemButton ) then
					block.itemButton:ClearAllPoints()
					if onLeft then
						block.itemButton:SetPoint("TopLeft", block, -28, -17)
					else
						block.itemButton:SetPoint("TopRight", block, -2, -17)
					end
				end
			end
		end
	end	
end

function handler:RepositionTracker()
	if not self:GetScroll() then
	    --Make the ObjectiveTrackerFrame display as normal.
		self.tracker:ClearAllPoints()
		self.tracker:SetPoint('TopLeft', self, 30, 0)
		self.tracker:SetPoint('TopRight', self, 0, 0)
		self.tracker:SetPoint('Bottom', self, 0, 0)
		self.heightSet = nil
		return
	elseif not self.heightSet then
	    --makes the ObjectiveTrackerFrame long enough to display ALL tracked quests for se in scrolling.
		self.tracker:ClearAllPoints()
		self.heightSet = true
		self.tracker:SetPoint('TopLeft', self, 30, 0)
		self.tracker:SetPoint('TopRight', self, 0, 0)
		self.tracker:SetHeight(2000)
	end
	self:UpdateItemButtons()
end

function handler:GetSettings()
    --settings. plan to make a stand alone version of the handler.
    return self.sets or unknown
end

local maxDisplayed, filterByZone, proximitySort, enableScroll, onLeft = 25
--defaults

do  --Rewrite specific functions used in  the QUEST_TRACKER_MODULE to enable quest sorting.
	local questList = {}
	local oldGetNumQuestWatches = GetNumQuestWatches
	local oldGetQuestWatchInfo = GetQuestWatchInfo
	
	function GetQuestWatchInfo(...)
		if handler:GetQuestsHidden() ~= true then
			if questList[...] then
				return unpack(questList[...])
			end
		end
	end
	
	local function POIdataExists()
	    --checks current zone for zone specific quests.
	    --if no zone specific quests, we display all quests instead.
		if handler:GetFilterByZone() == true then
			for i = 1, oldGetNumQuestWatches() do
				local _, _, _, _, _, _, _, _, _, _, _, _, _, isOnMap, hasLocalPOI = oldGetQuestWatchInfo(i)
				if (isOnMap == true)  or (hasLocalPOI == true)then
					return true
				end
			end
		end
	end

	function GetNumQuestWatches(...)
		if handler:GetQuestsHidden() == true then
			return 0
		end
		wipe(questList)
		local hasPoi = POIdataExists()
		for i = 1, oldGetNumQuestWatches() do
			local questInfo = {oldGetQuestWatchInfo(i)};
			questInfo.distance = GetDistanceSqToQuest(questInfo[3]) or 100000000
			if (hasPoi) then
				if (questInfo[14] == true) then
					tinsert(questList, questInfo)
				end
			else
				tinsert(questList, questInfo)
			end
		end

		if handler:GetSortByProximity() then
			table.sort(questList, function(a,b) return a.distance < b.distance  end);
		end

		return handler:GetMaxDisplayed()
	end
end

function handler:ToggleQuests()
    self.hideQuests = not self.hideQuests
end

function handler:GetQuestsHidden()

end

function handler:UpdateList()
	ObjectiveTracker_Update()
end

function handler:GetMaxDisplayed()
    local sets = self:GetSettings()
	if sets then
		return sets.maxDisplayed
	else
		return maxDisplayed
	end
end

function handler:SetMaxDisplayed(i)
    local sets = self:GetSettings()
	if sets then
		sets.maxDisplayed = i or maxDisplayed
	else
		maxDisplayed = i
	end
	handler:UpdateList()
end

function handler:GetFilterByZone()
    local sets = handler:GetSettings()
	if sets then
		return sets.filterByZone
	else
		return filterByZone
	end
end

function handler:ToggleFilterByZone()
    local sets = handler:GetSettings()
	if sets then
		sets.filterByZone = not sets.filterByZone
	else
		filterByZone = not filterByZone
	end
	handler:UpdateList()
end

function handler:GetSortByProximity()
    local sets = handler:GetSettings()
	if sets then
		return sets.proximitySort
	else
		return proximitySort
	end
end

function handler:ToggleSortByProximity()
    local sets = handler:GetSettings()
	if sets then
		sets.proximitySort = not sets.proximitySort
	else
		proximitySort = not proximitySort
	end
	handler:UpdateList()
end

function handler:GetScroll()
    local sets = handler:GetSettings()
	if sets then
		return sets.enableScroll
	else
		return enableScroll
	end
end

function handler:ToggleScroll()
    local sets = handler:GetSettings()
	if sets then
		sets.enableScroll = not sets.enableScroll
	else
		enableScroll = not enableScroll
	end
	handler:RepositionTracker()
	handler:UpdateList()
end

function handler:GetAlignment()
    local sets = handler:GetSettings()
	if sets then
		return sets.onLeft
	else
		return onLeft
	end
end

function handler:ToggleAlignment()
    local sets = handler:GetSettings()
	if sets then
		sets.onLeft = not sets.onLeft
	else
		onLeft = not onLeft
	end
	handler:UpdateItemButtons()
end

if not Dominos then
    handler:Initialize()
    return
end

	--[[ Globals ]]--
local AddonName = ...
local Addon = LibStub('AceAddon-3.0'):GetAddon("Dominos")
local KeyBound = LibStub('LibKeyBound-1.0')

--[[ Bar ]]--
local QuestBar = Addon:CreateClass('Frame', Addon.Frame)
QuestBar.master = Addon
_G[...] = QuestBar

do

	local QuestLogController = QuestBar.master:NewModule('QuestLogBar', 'AceEvent-3.0')

	function QuestLogController:Load()
		self.bar = QuestBar:New()
		self.bar:Layout()
	end

	function QuestLogController:Unload()
		self:UnregisterAllEvents()

		if self.bar then
			self.bar:Free()
		end
	end

	function QuestLogController:UpdateButtons()
		if InCombatLockdown() then return end
		self.bar:UpdateButtons()
	end


	function QuestBar:New()
		local frame = QuestBar.proto.New(self, 'quest')

			frame.sets.maxDisplayed = frame.sets.maxDisplayed or 25

		frame.scroll = handler:InitializeController(frame, frame.sets)
		return frame
	end

	function QuestBar:Layout()
		local w, h = ObjectiveTrackerFrame:GetSize()
		local pad = self:GetPadding()
		self:SetSize((175+pad)*1.5, (self.sets.height+pad)*1.5)
		self.sets.scrollStep = self.sets.scrollStep or 50
		self.scroll:ClearAllPoints()
		self.scroll:SetPoint("Center")
		self.scroll:SetSize((175)*1.5, (self.sets.height)*1.5)
		self.scroll:RepositionTracker()
	end

	function QuestBar:SetMaxDisplayed(value)
		self.sets.maxDisplayed = value
		handler:UpdateList()
	end	

	function QuestBar:SetFilterByZone(enable)
		self.sets.filterByZone = enable or nil
		handler:UpdateList()
	end

	function QuestBar:SetSortByProximity(enable)
		self.sets.proximitySort = enable or nil
		handler:UpdateList()
	end

	function QuestBar:onLeft(enable)
		self.sets.onLeft = enable or nil
		self.scroll:RepositionTracker()
	end
	
	function QuestBar:enableScroll(enable)
		self.sets.enableScroll = enable or nil
		self.scroll:RepositionTracker()
	end
	
	function QuestBar:GetDefaults()
		return {
			point = 'TopRight',
			spacing = 0,
			x = 0,
			y = -200,
			width = 232,
			height = 420,
			scrollStep = 50,
			enableScroll = true,
			pad = 0,
			maxDisplayed = 25,
		}
	end

	local L
	
	local function AddLayoutPanel(menu)
		local panel = menu:NewPanel(L.Layout)
		panel:NewSlider('Scroll Speed', 1, 250, 1,
			function(self) --Getter
				self:SetValue(self:GetParent().owner.sets.scrollStep)
			end,
			function(self) --Setter
				self:GetParent().owner.sets.scrollStep = self:GetValue()
			end
		)
		
		panel.opacitySlider = panel:NewOpacitySlider()
		panel.fadeSlider = panel:NewFadeSlider()
		panel.paddingSlider = panel:NewPaddingSlider()
		panel:NewSlider('Height', 250, 500, 1,
			function(self) --Getter
				self:SetValue(self:GetParent().owner.sets.height)
			end,
			function(self) --Setter
				self:GetParent().owner.sets.height = self:GetValue()
				self:GetParent().owner:Layout()
			end
		)
		panel:NewSlider('Width', 75, 250, 1,
			function(self) --Getter
				self:SetValue(self:GetParent().owner.sets.width)
			end,
			function(self) --Setter
				self:GetParent().owner.sets.width = self:GetValue()
				self:GetParent().owner:Layout()
			end
		)
	
		panel:NewSlider('Max Quests', 1, 25, 1,
			function(self) --Getter
				self:SetValue(self:GetParent().owner.sets.maxDisplayed)
			end,
			function(self) --Setter
				self:GetParent().owner:SetMaxDisplayed(self:GetValue())
			end
		)
	
		local zoneFilter = panel:NewCheckButton("Zone Filter")
		zoneFilter:SetScript('OnClick', function(self) self:GetParent().owner:SetFilterByZone(self:GetChecked()) end)
		zoneFilter:SetScript('OnShow', function(self) self:SetChecked(self:GetParent().owner.sets.filterByZone) end)
	
		local proximitySort = panel:NewCheckButton("Sort By Proximity")
		proximitySort:SetScript('OnClick', function(self) self:GetParent().owner:SetSortByProximity(self:GetChecked()) end)
		proximitySort:SetScript('OnShow', function(self) self:SetChecked(self:GetParent().owner.sets.proximitySort) end)
	
		local onLeft = panel:NewCheckButton("Buttons on Left")
		onLeft:SetScript('OnClick', function(self) self:GetParent().owner:onLeft(self:GetChecked()) end)
		onLeft:SetScript('OnShow', function(self) self:SetChecked(self:GetParent().owner.sets.onLeft) end)
		
		local enableScroll = panel:NewCheckButton("Enable Scroll")
		enableScroll:SetScript('OnClick', function(self) self:GetParent().owner:enableScroll(self:GetChecked()) end)
		enableScroll:SetScript('OnShow', function(self) self:SetChecked(self:GetParent().owner.sets.enableScroll) end)
	
		panel.scaleSlider = panel:NewScaleSlider()
		return panel
	end
	
	local function AddShowState(self)
		local p = self:NewPanel(L.ShowStates)
		p.height = 56
		local editBox = CreateFrame('EditBox', p:GetName() .. 'StateText', p, 'InputBoxTemplate')
		editBox:SetWidth(148) editBox:SetHeight(20)
		editBox:SetPoint('TOPLEFT', 12, -10)
		editBox:SetAutoFocus(false)
		editBox:SetScript('OnShow', function(self)
			self:SetText(self:GetParent().owner:GetShowStates() or '')
		end)
		editBox:SetScript('OnEnterPressed', function(self)
			local text = self:GetText()
			self:GetParent().owner:SetShowStates(text ~= '' and text or nil)
			self:ClearFocus()
		end)
		editBox:SetScript('OnEditFocusLost', function(self) self:HighlightText(0, 0) end)
		editBox:SetScript('OnEditFocusGained', function(self) self:HighlightText() end)
		local set = CreateFrame('Button', p:GetName() .. 'Set', p, 'UIPanelButtonTemplate')
		set:SetWidth(30) set:SetHeight(20)
		set:SetText(L.Set)
		set:SetScript('OnClick', function(self)
			local text = editBox:GetText()
			self:GetParent().owner:SetShowStates(text ~= '' and text or nil)
			editBox:SetText(self:GetParent().owner:GetShowStates() or '')
			editBox:ClearFocus()
		end)
		set:SetPoint('BOTTOMRIGHT', -8, 2)
		return p
	end
	
	local function AddAdvancedPanel(menu)
		local panel = menu:NewPanel(L.Advanced)
		panel:NewClickThroughCheckbox()
		panel:NewShowInOverrideUICheckbox()
		panel:NewShowInPetBattleUICheckbox()
		return panel
	end
	
	function QuestBar:CreateMenu()
		local menu = Dominos:NewMenu(self.id)
		L = LibStub('AceLocale-3.0'):GetLocale('Dominos-Config')
	
		AddLayoutPanel(menu)
		AddShowState(menu)
		if self.components then
			for i, b in pairs(self.components) do
				if b.CreateMenu and QuestBar.Menu then
					b:CreateMenu(menu)
				end
			end
		end
	
		AddAdvancedPanel(menu)
		return menu
	end
	
	
	local menu, config
	function QuestBar:ShowMenu()
		if not QuestBar.master:IsConfigAddonEnabled() then return end
		config = config or LoadAddOn('Dominos_Config')
		L = L or LibStub('AceLocale-3.0'):GetLocale('Dominos-Config')
		QuestBar.menu = QuestBar.menu or self:CreateMenu()
	
		local menu = QuestBar.menu
		menu:Hide()
		menu:SetOwner(self)
		menu:ShowPanel(L.Layout)
		menu:Show()
	end
end
